/*
 * (C) 2003 Clemson University
 *
 * See COPYING in top-level directory.
 */

#include "coven.h"
#include <stdlib.h>
#include <string.h>

/* this returns a void * pointer to the buffer data, this is done because
 * we chose a rather odd type for the buffer data and it can be confusing */
void * COVEN_Buffer_Data(struct COVEN_Buffer *buf)
{
	return (void*)((unsigned int)(&(buf->data)));
}

/* put this buffer into the TPH */
/* add a new tag to the list of ones in the TPH */
void COVEN_Put_Buffer(struct COVEN_TPH *tph, struct COVEN_Buffer *buf)
{
	struct COVEN_Buffer_List_El *newel = 
	  (struct COVEN_Buffer_List_El*)COVEN_malloc(
	  sizeof(struct COVEN_Buffer_List_El));
	newel->buffer = buf;
	newel->next = NULL;
	
	if(!tph->head) /* empty TPH */
	    tph->head = newel; /* so this gets to be the head */
	else 
	    tph->tail->next = newel; /* tail now points to this one */
	tph->tail = newel; /* regardless, I'm the new tail */
}

void COVEN_Move_Buffer(void **bufdata, struct COVEN_TPH *oldtph, 
  struct COVEN_TPH *newtph)
{
	struct COVEN_Buffer_List_El *bufel = oldtph->head;

	while(bufel) {
	    /* look at this next line too long and blood will pour from
	     * your nose */
	    if(*bufdata == COVEN_Buffer_Data(bufel->buffer)) {
		/* found it, let's move the owner */
		/* was part of oldtph, now owned by newtph */
		bufel->buffer->tph = newtph; 
		/* but we gotta malloc a bufel struct real quick for the new 
		 * tph */
		COVEN_Put_Buffer(newtph, bufel->buffer);
		return;
	    }
	    bufel = bufel->next;
	}
}

/* COVEN_Buffer_Replace is a macro, Do_Buffer_Replace is what that calls.
 * This function assumes that the data passed in should be COPIED, not just
 * a pointer to it inserted.  Therefore, the calling function is in charge
 * of freeing it after a call to this.
 * Pass in new_data = NULL if you want this to not actually copy the data
 * over but to just replace the buffer */
struct COVEN_Buffer * COVEN_Do_Buffer_Data_Replace(char *local_binding, 
  struct COVEN_TPH *tph, struct COVEN_Mod_Binding_Table *btable, 
  void *new_data, int num_elements)
{
	char *global_binding;
	struct COVEN_Buffer *buf;
	struct COVEN_Buffer *newbuf = NULL;
	void *newbuf_data;
	int num_bytes;
	
	COVEN_debug(DTPH, "BUFFER_REPLACE: starting - LB = %s\n", 
	  local_binding);
	/* first, find the buffer that we're replacing */
	global_binding = COVEN_Get_Global_Binding(btable, local_binding);
	if(global_binding == NULL) return NULL;

	buf = COVEN_Get_Buffer(tph, global_binding, -1);

	/* create a new buffer w/ the same tag, type, and inside of
	 * the same TPH as the other one - just with a new number of
	 * elements */
	newbuf = COVEN_Just_Create_Buffer(tph, buf->tag, num_elements, 
	  buf->type);
	newbuf->seg_num = buf->seg_num;
	newbuf->num_segs = buf->num_segs;

	/* determine the number of bytes of the new buffer */
	if(buf->type < PSE_TYPE_OFFSET)
	    num_bytes = COVEN_get_type_size(buf->type);
	else
	    num_bytes = PSE_get_type_size(buf->type);

	num_bytes *= num_elements;
	newbuf_data = COVEN_Buffer_Data(newbuf);
	/* copy the new buffer data into the new buffer */
	if(new_data != NULL) {
	    memcpy(newbuf_data, new_data, num_bytes);
	    /* change the buffer element so that it now points to this
	     * new buffer */
	}

	COVEN_Do_Buffer_Replace(tph, buf, newbuf);
	
	return newbuf;
}

void COVEN_Do_Buffer_Replace(struct COVEN_TPH *tph, 
  struct COVEN_Buffer *orig_buf, struct COVEN_Buffer *new_buf)
{
	int found;
	struct COVEN_Buffer_List_El *bufel;
	
	/* now, find that buffer element in the TPH, so that we can
	 * replace it's buffer */
	bufel = tph->head;
	found = FALSE;
	while(bufel && !found) {
	    if(bufel->buffer == orig_buf) {
	        /* ok, we found it - now let's replace the buffer in there */
	    	found = TRUE;

		bufel->buffer = new_buf;

		/* free the old buffer, now that we're not using it */
		COVEN_Free_Buffer(orig_buf);
	    }
	    bufel = bufel->next;
	}
}
/* free's a buffer */
/* this is a function incase some day we change what it really means to
 * free a buffer, ie: more complicated */
void COVEN_Free_Buffer(struct COVEN_Buffer *buffer)
{
   	COVEN_debug(DTPH,"FreeBUFFER(): buffer at %x\n", (unsigned int)buffer);
   	COVEN_debug(DTPH,"FreeBUFFER(): tag = %s\n", buffer->tag);
	COVEN_free(buffer->tag);
	COVEN_debug(DTPH,"FreeBUFFER(): buffer tag freed.\n");
	COVEN_free(buffer);
	COVEN_debug(DTPH,"FreeBUFFER(): buffer is now freed!\n");
}

/* type coming in is TYPE_PTR-20 b/c we're creating a buffer of
 * TYPEs, not a buffer of TYPE_PTRs... subtle.
 * If the ptype is negative, then the caller doesn't want us to do a 
 * check if it's the right type so just check for tag */
struct COVEN_Buffer * COVEN_Get_Buffer(struct COVEN_TPH *tph,
  char *find_tag, int ptype)
{
	struct COVEN_Buffer_List_El *bufel = tph->head;

	/* look through the list of elements, which is setup like a 
	 * linked-list */
	while(bufel) {
	    /* if we find it, return a pointer to the buffer */
	    if(!strcmp(bufel->buffer->tag,find_tag)) {
	        if(ptype < 0) return bufel->buffer;
		else if(ptype == bufel->buffer->type) return bufel->buffer;
		else bufel = bufel->next;
	    }
	    /* else, check the next element */
	    bufel = bufel->next;
	}

	/* if we didn't find the tag, return NULL */
	return NULL;
}

/* psize is the number of elements
 * psize * COVEN_get_type_size(ptype) is the number of bytes
 * New_Buffer creates the buffer and puts it in the TPH - this is the
 * normal route. */
struct COVEN_Buffer * COVEN_New_Buffer(struct COVEN_TPH *tph,
  char *tag, int psize, int ptype)
{
	struct COVEN_Buffer *buf;
	
	buf = COVEN_Just_Create_Buffer(tph, tag, psize, ptype);
	COVEN_Put_Buffer(tph, buf);
	return buf;
}

/* Just_Create_Buffer only creates the buffer, the buffer "dangles"
 * with no buffer element holding it until it's been "Put" into the 
 * TPH.  Usually this function should be called only through New_Buffer
 */
struct COVEN_Buffer * COVEN_Just_Create_Buffer(struct COVEN_TPH *tph,
  char *tag, int psize, int ptype)
{
	struct COVEN_Buffer *buf;
	int size;

	/* determine which type this comes from */
	if(ptype < PSE_TYPE_OFFSET) size = COVEN_get_type_size(ptype);
	else size = PSE_get_type_size(ptype);
	
	buf = (struct COVEN_Buffer *)COVEN_malloc(sizeof(struct COVEN_Buffer)+
	  psize * size);
	if(!buf) {
	    perror("\tError allocing memory: "); 
	    printf("\tTried to malloc %d bytes of memory and got an error, "
	      "exiting\n",
	      sizeof(struct COVEN_Buffer)+psize * size);
	    MPI_Finalize();
	    exit(1);
	}
	buf->type = ptype;
	buf->size = psize;
	buf->tph = tph;
	buf->tag = (char*)COVEN_malloc((strlen(tag)+1)*sizeof(char));
	strcpy(buf->tag,tag);

	return buf;
}

void * COVEN_Get_Buffer_Data(struct COVEN_TPH *tph, char *tag,
  int ptype)
{
	struct COVEN_Buffer *buf;
	buf = COVEN_Get_Buffer(tph, tag, ptype);
	if(!buf) {
	    fprintf(stderr, "\tERROR: tag `%s' NOT FOUND.\n", tag);
	    return NULL;
	}
	else
	    return COVEN_Buffer_Data(buf);
}

int COVEN_Size(void *data) 
{
	int size;
	int type;

	type = ((struct COVEN_Buffer *)((unsigned char*)data-(sizeof(
	  struct COVEN_Buffer)-sizeof(int))))->type;

	if(type < PSE_TYPE_OFFSET) size = COVEN_get_type_size(type);
	else size = PSE_get_type_size(type);
	
	return ((struct COVEN_Buffer *)((unsigned char*)data-(sizeof(
	  struct COVEN_Buffer)-sizeof(int))))->size * size;
}

/* return #elements, not #bytes */
int COVEN_Count(void *data) 
{
	return ((struct COVEN_Buffer *)((unsigned char*)data-(sizeof(
	  struct COVEN_Buffer)-sizeof(int))))->size;
}
