/*
 * (C) 2003 Clemson University
 *
 * See COPYING in top-level directory.
 */

#include "coven.h"
#include <stdlib.h>
#include <string.h>
#include "medea.h"

void COVEN_Do_Create_Action(struct COVEN_Collective_Action *action,
  COVEN_Type datatype, char *local_binding, struct COVEN_TPH *tph,
  struct COVEN_Mod_Binding_Table *btable, COVEN_Collective_Type ctype)
{
	char *global_binding;

	global_binding = COVEN_Get_Global_Binding(btable, local_binding);
	if(global_binding == NULL) return;

	/* find the buffer using the global binding in the given TPH. 
	 * a -1 here for the type means we don't want a type check done,
	 * we do that because at this stage we don't really know the type */
	action->buffer = COVEN_Get_Buffer(tph, global_binding, -1);
	action->datatype = datatype;
	action->type = ctype;
}

/* this takes a TPH (copy_tph) which it uses to model each other TPH
 * after.  However, in that copy it excludes all buffers that are to
 * be scattered.  That's because CINT_Scatter (called here) will fill
 * in these buffers */
void COVEN_Do_Scatter(struct COVEN_TPH *copy_tph, int num_tphs, 
  struct COVEN_Collective_Action *actions, int num_actions)
{
	int i;
	
	COVEN_TPH *tphs;
	tphs = (COVEN_TPH*)COVEN_malloc(num_tphs*sizeof(COVEN_TPH));
	for(i=0; i<num_tphs; i++) {
	    tphs[i] = COVEN_Copy_TPH_Exclude(copy_tph, actions, num_actions);
	    COVEN_Print_TPH(tphs[i]);
	}

	for(i=0; i<num_actions; i++) {
	    CINT_Scatter(actions[i].datatype, num_tphs, tphs,
	      actions[i].buffer);
	}

	for(i=0; i<num_tphs; i++) {
	    COVEN_Print_TPH(tphs[i]);
	    COVEN_debug(DCOL, "Scatter: Relaying that TPH!  Here we GO!\n");
	    COVEN_Relay_TPH_Simple(tphs[i]);
	}
	COVEN_Schedule_TPH_For_Dealloc(copy_tph);
}

void COVEN_Do_Gather(struct COVEN_TPH *from_tph, struct COVEN_TPH *to_tph,
  int num_elements, struct COVEN_Collective_Action *actions, int num_actions)
{
   	struct COVEN_Buffer *gather_to_buffer;
	int seg_num, num_segs;
	/* we make this static so that it grows over time */
	static int num_gathered;

	if(num_actions > 1) {
	    COVEN_Error("COVEN_Do_Gather(): We cannot yet handle multiple "
	      "actions.  Please ask Nathan to fix this if you have found "
	      "a reason for this.\n");
	}

	COVEN_debug(DCOL, "GATHER: Gathering a buffer with a size of %d "
	  "elements.\n", actions[0].buffer->size);
	COVEN_debug(DCOL, "GATHER: %d gathered so far.\n", num_gathered);
	
	/* find the buffer in the static TPH to gather into */
	/* look for the same tag (global binding) in the TO_TPH as was
	 * in the FROM_TPH (which is the same TPH that we setup the
	 * collective action with. */
	COVEN_debug(DCOL, "GATHER: Looking for buffer with tag '%s' in "
	  "static TPH . . .\n", actions[0].buffer->tag);
	gather_to_buffer = COVEN_Get_Buffer(to_tph, actions[0].buffer->tag,
	  actions[0].buffer->type);
	/* if it's null, then we didn't find the buffer */
	if(gather_to_buffer == NULL) {
	    /* this must be the first time, let's create the buffer
	     * for the first time here */
	    COVEN_debug(DCOL, "GATHER: No previous buffer.  Creating now "
	      "with size = %d elements.\n", num_elements);
	    /* the size of this buffer is passed in (at least the number of
	     * elements) and we want it to be of the same type */
	    gather_to_buffer = COVEN_New_Buffer(to_tph, actions[0].buffer->tag,
	      num_elements, actions[0].buffer->type);
	}
	
	COVEN_debug(DCOL, "GATHER: I have the buffer now, at address %x\n",
	  (unsigned int)gather_to_buffer);
	
	/* the 2nd argument, the tph_count, to Gather doesn't really do
	 * anything - we must pass in exactly 1 TPH anyway */
	CINT_Gather(actions[0].datatype, 1, &from_tph, gather_to_buffer);

	/* determine which TPH in the sequence this was and keep this
	 * information together so that we know when to eject the TPH */
	seg_num = actions[0].buffer->seg_num;
	num_segs = actions[0].buffer->num_segs;
	
	num_gathered++;

	COVEN_debug(DCOL, "GATHER: done, by the way this was %d of %d bufs.\n",
	  seg_num, num_segs);

#if 0
	{
	    MEDEA_BODY_TYPE from = (MEDEA_BODY_TYPE)COVEN_Buffer_Data(
	      actions[0].buffer);
	    MEDEA_BODY_TYPE to = (MEDEA_BODY_TYPE)COVEN_Buffer_Data(
	      gather_to_buffer);
	    int i;
	    COVEN_debug(DCOL, "  GATHER: ** from **\n");
	    for(i=0; i<5; i++) {
		COVEN_debug(DCOL, "  GATHER FROM: body[%d] pos = "
		  "%3.3f, %3.3f, %3.3f\n", i, from[i].position.x,
		  from[i].position.y, from[i].position.z);
	    }
	    COVEN_debug(DCOL, "  GATHER: ** to ** (ALL THE ELEMENTS)\n");
	    for(i=0; i<20; i++) {
		COVEN_debug(DCOL, "    GATHER TO: body[%d] pos = "
		  "%3.3f, %3.3f, %3.3f\n", i, to[i].position.x,
		  to[i].position.y, to[i].position.z);
	    }
	}
#endif
	
	/* if this is the last TPH, then it's time to eject this TPH (the
	 * result of all the gathers) */
	if(num_gathered == num_segs) {
	    COVEN_debug(DCOL, "GATHER: I have gathered all TPHs!\n");
	    COVEN_Print_TPH(from_tph);
	    COVEN_Do_Buffer_Replace(from_tph, actions[0].buffer,
	      gather_to_buffer);
	    COVEN_debug(DCOL, "GATHER: Now, after the replacement of the "
	      "original buffer, here's our new TPH!\n");
	    COVEN_Print_TPH(from_tph);
	}
	
	else {
	    COVEN_Schedule_TPH_For_Dealloc(from_tph);
	}
}
