/*
 * (C) 2003 Clemson University
 *
 * See COPYING in top-level directory.
 */

#include "coven.h"
#include <dlfcn.h>
#include <stdlib.h>
#include <string.h>

/* the module structures */
int num_Master_Tasks, num_Slave_Tasks;
struct COVEN_Program *program;
struct COVEN_Consts *constants;
struct COVEN_Mods *modules;
struct COVEN_Mod_Calls *mod_calls;

struct COVEN_Interface **mod_iface;

struct COVEN_TPH *ctph; /* the constant TPH */

void **mod_dls;

extern int myMPIrank;

void * COVEN_Open_Dynamics(void)
{
	void * dl;
	dl = dlopen("./coven_program.so", RTLD_LAZY);
	if(!dl) COVEN_Error("dlopen() on coven_program.so");
	return dl;
}

void COVEN_Close_Dynamics(void *dl)
{
	dlclose(dl);
}

void COVEN_Load_Dynamics(void *dl, struct COVEN_Tag_Usage *tagusage)
{
	int i, j;
	
	/* dynamically load the coven program and hook up some helper pointer
	 * and variables for others to use */
	program = (struct COVEN_Program*)COVEN_dl_get(dl, "_coven_program");
	num_Master_Tasks = program->num_master_tasks;
	num_Slave_Tasks = program->num_slave_tasks;
	constants = program->constants;
	modules = program->modules;
	mod_calls = program->mod_calls;

	if(myMPIrank == 0) {
	    COVEN_Print_Program(program);
	}
	ctph = COVEN_Setup_Constant_TPH(constants);
	mod_dls = (void**)COVEN_malloc(modules->num_mod_decls*sizeof(void*));
	mod_iface = (struct COVEN_Interface**)COVEN_malloc(modules->num_mod_decls*
	  sizeof(struct COVEN_Interface*));
	for(i=0; i<modules->num_mod_decls; i++) {
	    mod_dls[i] = dlopen(modules->mod_decls[i].file_name, RTLD_LAZY);
	    if(!mod_dls[i]) { 
	    	fputs(dlerror(), stderr); MPI_Finalize(); exit(1); 
	    }
	    mod_iface[i] = (struct COVEN_Interface*)COVEN_dl_get(mod_dls[i], 
	      "_module_interface");
	    //trace1("IFACE @: %x\n", mod_iface[i]);
	    modules->mod_decls[i].interface = mod_iface[i];
	}

	for(i=0; i<mod_calls->num_mod_calls; i++) {
	    program->mod_calls->mod_calls[i].program_number = i;
	    /* find this module from the list we just loaded */
	    /* and when we find it, hook up the interface pointer from the 
	     * loaded module to this particular instance of the module call */
	    for(j=0; j<modules->num_mod_decls; j++) {
		if(!(strcmp(program->mod_calls->mod_calls[i].module_name, 
		  modules->mod_decls[j].module_name))) {
		    /* found it, copy the pointer */
		    program->mod_calls->mod_calls[i].interface = 
		      modules->mod_decls[j].interface;
		    break;
		}
	    }
	}
}

void * COVEN_dl_get(void *dl, char *symbol) 
{
	void * ret;
	char *error;

	ret = dlsym(dl, symbol);
	if((error = dlerror()) != NULL) {
	    fputs(error, stderr);
	    MPI_Finalize();
	    exit(1);
	}
	return ret;
}

struct COVEN_TPH * COVEN_Setup_Constant_TPH(struct COVEN_Consts *consts)
{
   	struct COVEN_TPH *atph;
	int i;
	/* tmp types, this is quicken than a cleaner way */
	char data_c;
	unsigned char data_uc;
	short data_s;
	unsigned short data_us;
	int data_i;
	unsigned int data_ui;
	float data_f;
	double data_d;

	atph = COVEN_New_TPH();
	
	for(i=0; i<consts->num_consts; i++) {
	    struct COVEN_Const_Decl *cd = &(consts->const_decls[i]);
	    struct COVEN_Attribute *new_at = NULL;

	    switch(cd->type) {
	    	case COVEN_CHAR:
		    data_c = cd->str_value[0];
		    new_at = COVEN_New_Attribute(cd->type, cd->global_binding,
		      &data_c);
		    break;
		case COVEN_UCHAR:
		    data_uc = cd->str_value[0];
		    new_at = COVEN_New_Attribute(cd->type, cd->global_binding,
		      &data_uc);
		    break;
		case COVEN_SHORT:
		    data_s = (short)(atoi(cd->str_value));
		    new_at = COVEN_New_Attribute(cd->type, cd->global_binding,
		      &data_s);
		    break;
		case COVEN_USHORT:
		    data_us = (unsigned short)(atoi(cd->str_value));
		    new_at = COVEN_New_Attribute(cd->type, cd->global_binding,
		      &data_us);
		    break;
		case COVEN_INT:
		    data_i = (int)(atoi(cd->str_value));
		    new_at = COVEN_New_Attribute(cd->type, cd->global_binding,
		      &data_i);
		    break;
		case COVEN_UINT:
		    data_ui = (unsigned int)(atoi(cd->str_value));
		    new_at = COVEN_New_Attribute(cd->type, cd->global_binding,
		      &data_ui);
		    break;
		case COVEN_FLOAT:
		    data_f = (float)(atof(cd->str_value));
		    new_at = COVEN_New_Attribute(cd->type, cd->global_binding,
		      &data_f);
		    break;
		case COVEN_DOUBLE:
		    data_d = (double)(atof(cd->str_value));
		    new_at = COVEN_New_Attribute(cd->type, cd->global_binding,
		      &data_d);
		    break;
		case COVEN_STRING:
		    new_at = COVEN_New_Attribute(cd->type, cd->global_binding,
		      cd->str_value);
		    break;
		default:
		    COVEN_Error("Corrupt data type");
	    }
	    COVEN_Attribute_Table_Append_Attrib(atph->at, new_at);
	}
	
	return atph;
}
