/*
 * (C) 2003 Clemson University
 *
 * See COPYING in top-level directory.
 */

#include "coven.h"
#include "mpi.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <signal.h>

int done = FALSE;

int myMPIrank;
int myCOVENTaskID = -1;
int myProcessorNumber;
int am_slave;
int num_Slave_Processors;

MPI_Comm MPI_COMM_THIS_PARALLEL_TASK;

/* the module structures */
extern int num_Master_Tasks, num_Slave_Tasks;

/* usage stuff */
struct COVEN_Tag_Usage *tagusage = NULL;

int main(int argc, char **argv)
{
	MPI_Group comm_world_group, SLAVE_GROUP;
	char processor_name[MPI_MAX_PROCESSOR_NAME];
	int numprocs, myrank, namelen, *master_procs, i;
	void *dl;
	double starttime;

	MPI_Init(&argc, &argv);

	MPI_Comm_size(MPI_COMM_WORLD, &numprocs);
	MPI_Comm_rank(MPI_COMM_WORLD, &myrank);
	myMPIrank = myrank;
	MPI_Get_processor_name(processor_name, &namelen);

	if(numprocs < 2) { 
	    fprintf(stderr, "ERROR: There must be at least 2 tasks.\n");
	    MPI_Finalize();
	    exit(1);
	}

	COVEN_Profile_Disable();
	gossip_enable_stderr();
	//gossip_set_debug_mask(1, DLOOP | DMODULE | DMAIN | DCTRL | DTPH | DARGS);
	gossip_set_debug_mask(1, DRESULTS | DSPECIAL);
	
//	| DMAIN | DRESULTS | DCTRL | DQUEUE |
//	  DMODULE | DTPH);

	COVEN_debug(DMAIN, "---[Process %d on %s]---\n", myrank, 
	  processor_name);
	COVEN_debug(DSPECIAL, "---[Process %d on %s]---\n", myrank, 
	  processor_name);

	dl = COVEN_Open_Dynamics();
	COVEN_Load_Dynamics(dl, tagusage);

	COVEN_debug(DMAIN,"# master tasks = %d, # slave tasks = %d\n",
	  num_Master_Tasks, num_Slave_Tasks);

	num_Slave_Processors = (numprocs - num_Master_Tasks) / num_Slave_Tasks;

	if(myrank < num_Master_Tasks) {
	    myCOVENTaskID = myrank + 1;
	    myProcessorNumber = 0;
	    am_slave = FALSE;
	}
	else {
	    myCOVENTaskID = ((myrank - num_Master_Tasks) % num_Slave_Tasks) + 1;
	    myProcessorNumber = ((myrank - num_Master_Tasks) / 
	      num_Slave_Tasks) + 1;
	    am_slave = TRUE;
	}

	/* make a group just for all tasks on slave processors */
	MPI_Comm_group(MPI_COMM_WORLD, &comm_world_group);
	master_procs = (int *)COVEN_malloc(num_Master_Tasks*sizeof(int));
	for(i=0; i<num_Master_Tasks; i++) master_procs[i] = i;
	MPI_Group_excl(comm_world_group, num_Master_Tasks, master_procs, 
	  &SLAVE_GROUP);
	MPI_Comm_create(MPI_COMM_WORLD, SLAVE_GROUP, &SLAVE_COMM);
	MPI_Group_free(&SLAVE_GROUP);
	COVEN_free(master_procs);

	MPI_COMM_THIS_PARALLEL_TASK = MPI_COMM_WORLD; 
	/* just set it to something */
	if(myrank >= num_Master_Tasks) { /* all slave tasks */
	    /* and make a group just for instances of myself on other 
	     * processors */
	    int *ranks ;
	    MPI_Group PARALLEL_GROUP, slave_group;
	    int i;
		
	    ranks = (int*)COVEN_malloc(num_Slave_Processors*sizeof(int));
	    for(i=0; i<num_Slave_Processors; i++) {
		ranks[i] = (myCOVENTaskID - 1) + (num_Slave_Tasks * i);
	    }
		
	    MPI_Comm_group(SLAVE_COMM, &slave_group);
	    MPI_Group_incl(slave_group, num_Slave_Processors, ranks, 
	      &PARALLEL_GROUP);
	    MPI_Comm_create(SLAVE_COMM, PARALLEL_GROUP, 
	      &MPI_COMM_THIS_PARALLEL_TASK);
	    MPI_Group_free(&PARALLEL_GROUP);
	    COVEN_free(ranks);
	}

	COVEN_debug(DMAIN,"MPIrank %d and COVENTaskID %d -> on processor "
	  "%d\n", myMPIrank, 
	  myCOVENTaskID, myProcessorNumber);

	MPI_Barrier(MPI_COMM_WORLD);
	starttime = MPI_Wtime();

	if(myrank == 0) {
	    COVEN_debug(DMAIN,"Starting job - NOW!\n");
	    COVEN_Master_Inject();
	}
	COVEN_Program_Sequencer();

	COVEN_Close_Dynamics(dl);

	MPI_Barrier(MPI_COMM_WORLD);

	if(myMPIrank == 0) {
	    double endtime, diffsecs;
	    int mins, secs;

	    endtime = MPI_Wtime();
	    diffsecs = endtime-starttime;
	    mins = diffsecs / 60.0;
	    secs = (int)diffsecs % 60;
	    COVEN_debug(DRESULTS,"\t%.2f tot.secs - %3d:%02d mins:seconds\n", 
	      diffsecs, mins, secs);
	}

	COVEN_debug(DMAIN,"calling barrier: %d\n", myMPIrank); 
	MPI_Barrier(MPI_COMM_WORLD);
	MPI_Finalize();
	
	return 0;
}
