/*
 * (C) 2003 Clemson University
 *
 * See COPYING in top-level directory.
 */

#include "coven.h"
#include <errno.h>
#include <stdlib.h>

extern int myMPIrank, myCOVENTaskID, myProcessorNumber, am_slave;
extern MPI_Comm MPI_COMM_THIS_PARALLEL_TASK;

/* the module structures */
extern struct COVEN_Program *program;
extern struct COVEN_Consts *constants;
extern struct COVEN_Mods *modules;
extern struct COVEN_Mod_Calls *mod_calls;

extern struct COVEN_Interface **mod_iface;
extern void **mod_dls;

extern struct COVEN_Tag_Usage *tagusage;

extern int done;

/* the static TPH array for the module calls for this program sequencer */
struct COVEN_TPH **stph;
/* the constant TPH that we send into each module call */
extern struct COVEN_TPH *ctph;

/* this is the TPH queue for locally queued TPHs - TPHs in this queue
 * are coming from THIS SAME MPI TASK and therefore are not marshalled */
struct COVEN_TPH_Queue q;

void COVEN_Master_Inject(void)
{
	struct COVEN_TPH *new_tph;
	void *marshalled_TPH;
	int marshalled_TPH_size;
	
	new_tph = COVEN_New_TPH();
	COVEN_Profile_Stamp_TPH_Created(new_tph);
	/* this is where we used to fill the default TPH with the "constants"
	 * from the parameter file, we should now do something to fill it with
	 * the consts struct */
	/* COVEN_Fill_TPH_With_Symbols(new_tph); */

	COVEN_Marshall_TPH(new_tph, &marshalled_TPH, &marshalled_TPH_size);

	if(stph == NULL) COVEN_Malloc_Static_TPHs();
	COVEN_process_received(new_tph);

	COVEN_free(marshalled_TPH);
}

void COVEN_Shut_Down(void)
{
	int i;
	int numprocs;

	COVEN_debug(DMAIN, "Shut_Down(): shutting down everything!\n");

	/* we barrier here because if there are other parallel tasks like
	 * us executing the same code we need to let them finish before we
	 * shut down.  For the most part, this will only be executed by the
	 * last module in the program and that should be a master task,
	 * which has only itself in this MPI_Comm */

	if(am_slave) {
	    COVEN_debug(DMAIN, "SHUTDOWN: Barriering....\n");
	    MPI_Barrier(MPI_COMM_THIS_PARALLEL_TASK);
	    COVEN_debug(DMAIN, "SHUTDOWN: ok, shutting down ....\n");
	}

	MPI_Comm_size(MPI_COMM_WORLD, &numprocs);

	/* send a message to each MPI task telling it to shut down */
	for(i=0; i<numprocs; i++) {
	    unsigned char byte = COVEN_SHUTDOWN_MESSAGE;
	    MPI_Request request;

	    COVEN_debug(DCTRL,"Telling MPItask %d to shutdown.\n", i);
	    MPI_Isend(&byte, 1, MPI_BYTE, i, COVEN_PROGRAM_SEQUENCER_TAG,
	      MPI_COMM_WORLD, &request);
	}

	for(i=0; i<mod_calls->num_mod_calls; i++) {
	    COVEN_Free_TPH(stph[i]);
	}
	COVEN_free(stph);
}

void COVEN_Program_Sequencer(void)
{
	int done = FALSE;
	void *recv_buffer;
	int num_TPHs_recvd;

	num_TPHs_recvd = 0;

	if(stph == NULL) COVEN_Malloc_Static_TPHs();
	
	while(!done) {
	    int count_received;
	    MPI_Status probe_status, status;
	    struct COVEN_TPH *tph;
	    int queued_TPH = FALSE;
	    
	    /* wait until someone sends us a TPH to process, sleep until then */
	    COVEN_debug(DQUEUE,"MPI%d,P%d,CTID%d: going to sleep on a "
	      "probe.\n", myMPIrank, myProcessorNumber, myCOVENTaskID);
	    MPI_Probe(MPI_ANY_SOURCE, COVEN_PROGRAM_SEQUENCER_TAG, 
	      MPI_COMM_WORLD, &probe_status);
	    COVEN_debug(DQUEUE,"\tMPI%d,P%d,CTID%d: WOKE UP FROM PROBE.\n", 
	      myMPIrank, myProcessorNumber, myCOVENTaskID);
	    
	    COVEN_debug(DQUEUE,"\t\tMPI%d,P%d,CTID%d: MESSage arrived from "
	      "MPI task %d\n", myMPIrank, myProcessorNumber, myCOVENTaskID, 
	      probe_status.MPI_SOURCE);
	    MPI_Get_count(&probe_status, MPI_BYTE, &count_received);
	    
	    COVEN_debug(DQUEUE,"\t\tMPI%d,P%d,CTID%d: %d bytes coming "
	      "in!!!!!\n", myMPIrank, myProcessorNumber, myCOVENTaskID, 
	      count_received);
	    
	    recv_buffer = (void*)COVEN_malloc(count_received);
	    MPI_Recv(recv_buffer, count_received, MPI_BYTE, 
	      probe_status.MPI_SOURCE, probe_status.MPI_TAG, MPI_COMM_WORLD, 
	      &status);
	    
	    /* it's a message - for ME!! */
	    if(count_received == 1) {
	        COVEN_debug(DQUEUE,"\t\tMPI%d,P%d,CTID%d: message %d!!!!!!\n",
		  myMPIrank, myProcessorNumber, myCOVENTaskID, 
		  ((unsigned char*)recv_buffer)[0]);
		if(((unsigned char*)recv_buffer)[0] == COVEN_SHUTDOWN_MESSAGE) {
		    done = TRUE;
		}
		else if(((unsigned char*)recv_buffer)[0] == 
		  COVEN_TPH_QUEUED_MESSAGE) {
		    queued_TPH = TRUE;
		}
	    }

	    if(count_received != 1 || queued_TPH == TRUE) {
	    	if(queued_TPH == TRUE) {
		    queued_TPH = FALSE;
		    tph = COVEN_Dequeue_TPH();
	 	}
	        else {
		    tph = COVEN_Unmarshall_TPH(recv_buffer);
	    	}
	    	/* now we have a TPH to work with */
	    	num_TPHs_recvd ++;

	    	/* now process the TPH we received */
	    	COVEN_process_received(tph);
	    }
	    /* however this receive buffer was used, free it now */
	    COVEN_free(recv_buffer);
	/* and then reloop, waiting for another message */
	}
}

void COVEN_process_received(struct COVEN_TPH *tph)
{
	/* COVEN_Print_TPH(tph); */
	struct COVEN_Attribute *at;
	int tph_dealloced, tph_manually_relayed, stepnum;
	/* by default, start at the first module to try and process */
	int start_module = 0; 

	COVEN_debug(DMAIN,"START -> PGM SEQ: TPH (%x)\n", (unsigned int)tph);
	COVEN_debug(DCTRL,"PROGRAM SEQUENCER: starting process TPH\n");
	COVEN_TPH_Clear_Manual_Relay(tph);
	tph_dealloced = FALSE;
	tph_manually_relayed = FALSE;

	/* if the target module number is set, then someone wants this to 
	 * goto a specific module inside of this program sequencer, not just
	 * the first module */
	at = COVEN_Attribute_Table_Find_Attrib(tph->at, 
	  "_coven_target_module_number");
	if(at != NULL) {
	    start_module = *(int*)COVEN_Attribute_Table_Get_Attrib_Data(tph->at,
	      "_coven_target_module_number");
	    /* clear this module target information b/c we don't want it to
	     * propogate further on */
	    COVEN_Attribute_Table_Delete_Attrib(tph->at, at);
	}

	/* loop through all the modules that I am supposed to run */
	for(stepnum=start_module; stepnum<mod_calls->num_mod_calls && 
	  !tph_dealloced && !tph_manually_relayed; stepnum++) {

	    /* if this module runs under me */
	    if((mod_calls->mod_calls[stepnum].slave_module == am_slave) && 
	      mod_calls->mod_calls[stepnum].task_number == myCOVENTaskID) {
	        COVEN_debug(DMODULE, "\t[MPI%d %c%d]: step %d -> %s\n", 
		  myMPIrank, am_slave ? 'S' : 'M', myCOVENTaskID, stepnum, 
		  mod_calls->mod_calls[stepnum].module_name);
		
		/* check for a loop instruction, only run it though before the
		 * module is this module is the begin module, we will also check
		 * after the module incase this is the end module */
		if(mod_calls->mod_calls[stepnum].loop != NULL &&
		   mod_calls->mod_calls[stepnum].loop->begin_module == stepnum) {
		    /* this loop may cause the TPH to relay back to the start
		     * of the loop, or perhaps to the end of the loop.  if so,
		     * we need to know this so that we don't send this TPH into
		     * the next module */
		    tph_manually_relayed = COVEN_Process_Loop(
		      mod_calls->mod_calls[stepnum].loop, tph, ctph, stepnum);
		}
	
		if(!tph_manually_relayed) {
		    COVEN_Profile_Stamp_TPH_Module_In(tph, 
		      &mod_calls->mod_calls[stepnum]);
		    COVEN_debug(DMODULE,"\t[MPI%d %c%d]: step %d -> %s "
		      "(STARTING)\n", 
		      myMPIrank, am_slave ? 'S' : 'M', myCOVENTaskID, stepnum,
		      mod_calls->mod_calls[stepnum].module_name);

		    /* actually run this module on the received TPH */
		    COVEN_Run_Module(tph, ctph, stph[stepnum],
		      &mod_calls->mod_calls[stepnum]);

		    COVEN_debug(DMODULE,"\t[MPI%d %c%d]: step %d -> %s "
		      "(DONE)\n", myMPIrank, 
		      am_slave ? 'S' : 'M', myCOVENTaskID, stepnum, 
		      mod_calls->mod_calls[stepnum].module_name);
		    
		    COVEN_Print_TPH(tph);

		    COVEN_Profile_Stamp_TPH_Module_Out(tph, 
		      &mod_calls->mod_calls[stepnum]);

		    /* check for a loop instruction */
		    if(mod_calls->mod_calls[stepnum].loop != NULL &&
		       mod_calls->mod_calls[stepnum].loop->end_module == 
		       stepnum) {
		    	/* this loop may cause the TPH to relay back to the 
			 * start of the loop, or perhaps to the end of the 
			 * loop.  if so, we need to know this so that we 
			 * don't send this TPH into the next module */
		    	tph_manually_relayed = COVEN_Process_Loop(
		      	  mod_calls->mod_calls[stepnum].loop, tph, ctph, 
			  stepnum);
		    }
		}
	
		/* we need to check here again for relaying because that module
		 * may have done some manual relaying!  Check first that we
		 * havn't already relayed it ourselves, then check if that last
		 * module did it */
		if(!tph_manually_relayed && COVEN_TPH_Manually_Relayed(tph)) {
		    tph_manually_relayed = TRUE;
		}
			
		if(COVEN_TPH_Scheduled_For_Dealloc(tph)) {
		    COVEN_debug(DMAIN,"\t[MPI%d %c%d]: TPH has been scheduled "
		      "for dealloc, doing so.\n", myMPIrank, 
		      am_slave ? 'S' : 'M', myCOVENTaskID);
		    COVEN_Profile_Stamp_TPH_Destroyed(tph);
		    COVEN_Free_TPH(tph);
		    COVEN_debug(DMAIN, "  PGM SEQ: freeing TPH\n");
		    tph_dealloced = TRUE;
		}
		/* else, let's cleanup any thing we can dealloc now */
		else if(!tph_manually_relayed) {
		    COVEN_debug(DTPH,"cleaning up\n");
		    COVEN_Cleanup_TPH(tph, stepnum, tagusage);
		    COVEN_debug(DTPH,"CLEANED up\n");
		}
	    }
	}
	/* determine which task runs next, this routine puts attribs
	 * into the TPH specifying the answer */
	if(!tph_dealloced && !tph_manually_relayed) {
	    int proc_type, task_id;
	
	    COVEN_debug(DMAIN, "PGMSQNCR: Trying to relay this through "
	      "GetNextTask.\n");
	    COVEN_Get_Next_Task(tph, &proc_type, &task_id);
	    /* if Get_Next_Task returned -1, -1 then it's b/c it called for a 
	     * shutdown which of course means we have nothing left to relay! */
	    if(proc_type != -1 && task_id != -1) {
	        COVEN_debug(DTPH,"PGMSQNCR: Relaying TPH to proc %d id %d\n",
		  proc_type, task_id);
		COVEN_Relay_TPH(tph, proc_type, -1, task_id);
		COVEN_debug(DTPH,"PGMSQNCR: Freeing TPH!\n");
		COVEN_debug(DTPH,"PGMSQNCR: Here's TPH first...\n");
		COVEN_Print_TPH(tph);
		COVEN_Free_TPH(tph);
		COVEN_debug(DMAIN, "  PGM SEQ: freeing TPH\n");
		COVEN_debug(DTPH,"PGMSQNCR: Done freeing TPH!\n");
		tph_dealloced = TRUE;
	    }
	}
	/* we have to be careful here, if the TPH has been relayed it may have
	 * been relayed to another processor, in which case the TPH will be
	 * marked "scheduled for dealloc".  If it was relayed to a module
	 * running in this same process, then it is relayed locally and
	 * not freed */
	COVEN_debug(DCTRL,"PROGRAM SEQUENCER: leaving process TPH\n");
	COVEN_debug(DMAIN,"EXIT -> PGM SEQ: TPH (%x) : REL %d DEALLOC %d\n", 
	  (unsigned int)tph, tph_manually_relayed, tph_dealloced);
}

void COVEN_Get_Next_Task(struct COVEN_TPH *tph, int *proc_type, int *task_id)
{
	int i;

	for(i=0; i<mod_calls->num_mod_calls; i++) {
	    if((mod_calls->mod_calls[i].slave_module == am_slave) && 
	      (mod_calls->mod_calls[i].task_number == myCOVENTaskID)) {
		/* if we're not at the end of the array */
		if(i != (mod_calls->num_mod_calls)-1) { 
		    /* if the next task is NOT the same task as me */
		    if(mod_calls->mod_calls[i+1].task_number > myCOVENTaskID ||
		      mod_calls->mod_calls[i+1].slave_module != am_slave) {
			/* then that's what task this data goes to! */
		        COVEN_debug(DCTRL,
			  "%c:%d -> %c:%d\n", am_slave ? 'S' : 'M',
			  myCOVENTaskID, 
			  mod_calls->mod_calls[i+1].slave_module ? 'S' : 'M',
			  mod_calls->mod_calls[i+1].task_number);
	
			*proc_type = (mod_calls->mod_calls[i+1].slave_module);
			*task_id = (mod_calls->mod_calls[i+1].task_number);
					
			return;
		    }
		}
		else {
		    COVEN_Shut_Down();
		    *proc_type = -1;
		    *task_id = -1;
	 	    return;
		}
	    }
	}
}

int COVEN_Run_Module(struct COVEN_TPH *tph, struct COVEN_TPH *ctph,
  struct COVEN_TPH *stph, struct COVEN_Mod_Call *mcall) 
{
	int retval;
	//COVEN_Start_Module(mcall);
	COVEN_debug(DMODULE,
	  "-] MODULE (calling: %s)\n", mcall->module_name);
	COVEN_Print_TPH(tph);
	COVEN_debug(DMODULE, "-] MODULE %s - here GOES!\n", mcall->module_name);
	retval = (*mcall->interface->func)(tph, ctph, stph, mcall->btable);
	COVEN_debug(DMODULE,
	  "-] MODULE (returned from: %s)\n", mcall->module_name);
	return retval;
}

#if 0
void COVEN_Start_Module (struct COVEN_Mod_Call *mcall) 
{
	/*
	mcall->next_arg = mcall->arguments;
	*/
}
#endif

/* prints out information on a module, including its interface */
void COVEN_Print_Module(struct COVEN_Mod_Call *mod) 
{
	COVEN_Print_Mod_Call(mod);
}

void COVEN_Malloc_Static_TPHs(void)
{
   	int i;

	/* here we allocate static TPHs for the modules that this program
	 * sequencer will control */
	/* since we have many module calls but this program sequencer only
	 * handles a subset of those calls we allocate space for all the TPHs
	 * and only use those TPHs for this program sequencer's module calls.
	 * The size of the TPH data struct is itself very small with just
	 * a few pointers this seems ok. */
	COVEN_debug(DTPH, "Mallocing static TPHs!\n");
	COVEN_debug(DTPH, "STPH Malloc: # (%d) * size (%d)\n",
	  mod_calls->num_mod_calls, sizeof(struct COVEN_TPH *));

	/* malloc the pointers to the TPHs */
	stph = (struct COVEN_TPH **)COVEN_malloc(mod_calls->num_mod_calls * 
	  sizeof(struct COVEN_TPH *));

	for(i=0; i<mod_calls->num_mod_calls; i++) {
	    stph[i] = COVEN_New_TPH();
	}

	COVEN_debug(DTPH, "Done mallocing static TPHs!\n");
}

/* returns true if this loop caused the TPH to move away from this module, such
 * as looping to the end or beginning of the loop.  This should cause the
 * caller to augment what he does with this TPH (such as not use it anymore) */
int COVEN_Process_Loop(struct COVEN_Loop *loop, struct COVEN_TPH *tph,
  struct COVEN_TPH *ctph, int this_mod_number)
{
	struct COVEN_Attribute *attrib;
	int cur_val = -1;
	int *cur_val_ptr = NULL;
	int cmp_val = -1;
	int relay_mod = -1;

	COVEN_debug(DLOOP, "LOOP: starting a processing seg.\n");

	/* first try and find the loop tag in the TPH */
	attrib = COVEN_Attribute_Table_Find_Attrib(tph->at, loop->loop_tag);

	/* if the attribute didn't exist, then this must be the first
	 * iteration, we should create it and initialize it */
	if(attrib == NULL) {
	    /* what if this is the end of the loop and this tag hasn't been
	     * created?  That would be very bad! */
	    if(this_mod_number == loop->end_module) {
	    	COVEN_Error("Loop tag did not exist at end of loop!\n");
	    }

	    cur_val_ptr = (int *)COVEN_Attribute_Table_Put_Attrib_Data(
	      tph->at, loop->loop_tag, COVEN_INT, 1);
	    *cur_val_ptr = loop->initial_val;
	    cur_val = *cur_val_ptr;
	}
	/* the attribute value existed, lets just grab the value */
	else {
	    cur_val_ptr = (int *)(attrib->data);
	    cur_val = *cur_val_ptr;
	}
	/* now we have the value of the loop variable in cur_val */
	COVEN_debug(DLOOP, "LOOP: current loop variable value = %d\n",
	  cur_val);
	COVEN_debug(DMEM, "LOOP: current loop variable value = %d\n",
	  cur_val);

	/* now, find the value of the comparison tag */
	attrib = COVEN_Attribute_Table_Find_Attrib(tph->at, loop->cmp_tag);
	/* if we didn't find it in the default TPH, let's check in the constant
	 * TPH */
	if(attrib == NULL) {
	    attrib = COVEN_Attribute_Table_Find_Attrib(ctph->at, loop->cmp_tag);
	    if(attrib == NULL) {
	    	/* ok, we still didn't find it - something's really wrong */
	        char str[128];
		snprintf(str, 128, "Loop LEQ tag '%s' not found in default or "
		  "constant TPH, don't know what to do.", loop->cmp_tag);
	        COVEN_Error(str);
	    }
	}
	/* attrib must be valid here, else we would have exited */
	cmp_val = *((int *)(attrib->data));
	COVEN_debug(DLOOP, "LOOP: cmp loop value = %d\n", cmp_val);
	
	/* ok, now we have the current value in 'cur_val' and the cmp value
	 * in 'cmp_val'. */

	/* now, unlike a "regular" loop we're going to check at the END of the
	 * loop to see if, after incrementing the loop counter jumping back to
	 * the start of the loop would just make us jump OUT of the loop.  We
	 * check here, rather than just passing it to the top and making him
	 * jump because we are concerned about performance and the overhead of
	 * scheduling this TPH on another process/(or) just to have him send it
	 * back because it was outside of the loop bounds */
	if(this_mod_number == loop->end_module) {
	    /* we must be the end of the loop */

	    /* increment the loop variable by the incremental value.  Since
	     * it's a pointer to the int value it automagically forces the int
	     * value "into" the TPH */
	    *cur_val_ptr += loop->incr_val;
	    cur_val = *cur_val_ptr;

	    COVEN_debug(DLOOP, "LOOP: Loop var is now %d.\n", *cur_val_ptr);
	    COVEN_debug(DMEM, "LOOP: Loop var is now %d.\n", *cur_val_ptr);
	}

	/* this handles all the loop types and does the comparison operations
	 * on them */
	switch(loop->cmp_type) {
	    case COVEN_LT:
	        /* loop->module_number is the module which has the ending of 
		 * the loop on it.  We want to loop to that module PLUS ONE 
		 * - just like a loop in assembly jumping to the "next 
		 * instruction". */
		if(cur_val >= cmp_val) {
		    /* we have to first delete the looping variable attrib
		     * so that when (if) we come back to this loop this
		     * variable will not already be set and already be too
		     * large!  Basically, this makes the loop look like:
		     * for(int i=....)
		     * instead of
		     * for(i=....)
		     * So, we have to create and delete the loop variable at
		     * the end of the looping scope.  This is done below in all
		     * the loop exiting conditions. */
		    struct COVEN_Attribute *a;
		    relay_mod = (loop->end_module)+1;
		    a = COVEN_Attribute_Table_Find_Attrib(tph->at, 
		      loop->loop_tag);
		    COVEN_Attribute_Table_Delete_Attrib(tph->at,
		      a);
		}
		else if(cur_val < cmp_val && 
		  this_mod_number == loop->end_module) {
		    relay_mod = loop->begin_module;
		}
		else return FALSE;
		COVEN_debug(DLOOP, "LOOP: LT relaying to module %d\n",
		  relay_mod);
		COVEN_Manually_Relay(tph, relay_mod);
		return TRUE; /* TRUE: this was manually relayed */
	    	break;
	    case COVEN_LEQ:
		if(cur_val > cmp_val) {
		    struct COVEN_Attribute *a;
		    relay_mod = (loop->end_module)+1;
		    a = COVEN_Attribute_Table_Find_Attrib(tph->at, 
		      loop->loop_tag);
		    COVEN_Attribute_Table_Delete_Attrib(tph->at,
		      a);
		}
		else if(cur_val <= cmp_val && 
		  this_mod_number == loop->end_module) {
		    relay_mod = loop->begin_module;
		}
		else return FALSE; /* FALSE: we don't want to relay this */
		
		/* if we're here, we must need to relay this somehow */
		COVEN_debug(DLOOP, "LOOP: LEQ relaying to module %d\n",
		  relay_mod);
		COVEN_Manually_Relay(tph, relay_mod);
		return TRUE; /* TRUE: this was manually relayed */
	    	break;
	    case COVEN_GT:
		if(cur_val <= cmp_val) {
		    struct COVEN_Attribute *a;
		    relay_mod = (loop->end_module)+1;
		    a = COVEN_Attribute_Table_Find_Attrib(tph->at, 
		      loop->loop_tag);
		    COVEN_Attribute_Table_Delete_Attrib(tph->at,
		      a);
		}
		else if(cur_val > cmp_val &&
		  this_mod_number == loop->end_module) {
		    relay_mod = loop->begin_module;
		}
		else return FALSE;
		COVEN_debug(DLOOP, "LOOP: GT relaying to module %d\n",
		  relay_mod);
		COVEN_Manually_Relay(tph, relay_mod);
		return TRUE; /* TRUE: this was manually relayed */
	    	break;
	    case COVEN_GEQ:
		if(cur_val < cmp_val) {
		    struct COVEN_Attribute *a;
		    relay_mod = (loop->end_module)+1;
		    a = COVEN_Attribute_Table_Find_Attrib(tph->at, 
		      loop->loop_tag);
		    COVEN_Attribute_Table_Delete_Attrib(tph->at,
		      a);
		}
		else if(cur_val >= cmp_val &&
		  this_mod_number == loop->end_module) {
		    relay_mod = loop->begin_module;
		}
		else return FALSE;
		COVEN_debug(DLOOP, "LOOP: GEQ relaying to module %d\n",
		  relay_mod);
		COVEN_Manually_Relay(tph, relay_mod);
		return TRUE; /* TRUE: this was manually relayed */
	    	break;
	    case COVEN_EQ:
		if(cur_val != cmp_val) {
		    struct COVEN_Attribute *a;
		    relay_mod = (loop->end_module)+1;
		    a = COVEN_Attribute_Table_Find_Attrib(tph->at, 
		      loop->loop_tag);
		    COVEN_Attribute_Table_Delete_Attrib(tph->at,
		      a);
		}
		else if(cur_val == cmp_val &&
		  this_mod_number == loop->end_module) {
		    relay_mod = loop->begin_module;
		}
		else return FALSE;
		COVEN_debug(DLOOP, "LOOP: EQ relaying to module %d\n",
		  relay_mod);
		COVEN_Manually_Relay(tph, relay_mod);
		return TRUE; /* TRUE: this was manually relayed */
	    	break;
	    case COVEN_NEQ:
		if(cur_val == cmp_val) {
		    struct COVEN_Attribute *a;
		    relay_mod = (loop->end_module)+1;
		    a = COVEN_Attribute_Table_Find_Attrib(tph->at, 
		      loop->loop_tag);
		    COVEN_Attribute_Table_Delete_Attrib(tph->at,
		      a);
		}
		else if(cur_val != cmp_val &&
		  this_mod_number == loop->end_module) {
		    relay_mod = loop->begin_module;
		}
		else return FALSE;
		COVEN_debug(DLOOP, "LOOP: NEQ relaying to module %d\n",
		  relay_mod);
		COVEN_Manually_Relay(tph, relay_mod);
		return TRUE; /* TRUE: this was manually relayed */
	    	break;
	    default: /* wtf? this isn't a loop comparison */
		COVEN_Error("Fatal error: unknown loop comparison type.\n");
		break;
	}
	
	/* if we got down here, then we must have met whatever requirements 
	 * of the loop to stay within the loop */
	return FALSE; /* FALSE: didn't manually relay this TPH in this loop */
}
